# -*- coding: utf-8 -*-
import json
import uuid
from unittest.mock import patch

from odoo import fields
from odoo.exceptions import ValidationError
from odoo.tests.common import tagged, TransactionCase
from odoo.addons.mail.tests.common import MockEmail
from odoo.addons.saas_worker.const import SAAS_BASE_URL, TEST
from odoo.addons.openerp_enterprise.tests.common import EnterpriseDatabaseCommon
from odoo.addons.openerp_enterprise.models.gandi_api import ContactType, TLDS, GANDI
from odoo.addons.openerp_enterprise.models.domain import PSL

from freezegun import freeze_time


class TestDNSSetup(TransactionCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.company = cls.env.ref('base.main_company')

        groups = [
            # Portal User
            cls.env.ref('base.group_portal').id,
        ]

        cls.input_data = {
            'fname': 'foo',
            'lname': 'bar',
            'email': 'portal_domain_owner@odoo.com',
            'phone': '0032490000000',
            'country': 'BE',
            'address': 'addr',
            'city': 'city',
            'zip': '1000',
        }

        cls.owner = cls.env['res.users'].with_context({'no_reset_password': True}).create({
            'name': '...',
            'login': 'portal_domain_owner',
            'email': cls.input_data['email'],
            'company_id': cls.company.id,
            'groups_id': [(6, 0, groups)],
        })

        cls.odoo_sa = {
            'type': 1,
            'orgname': 'Odoo SA',
            'given': 'Fabien',
            'family': 'Pinckaers',
            'email': 'gandi@odoo.com',
            'phone': '+32 (0)81 81 37 00',
            'country': 'BE',
            'streetaddr': 'Chaussée de Namur, 40',
            'city': 'Grand-Rosière',
            'zip': '1367',
            'data_obfuscated': True,
            'mail_obfuscated': True,
            'extra_parameters': {},
        }

        cls.sub = cls.env['sale.order']
        cls.db = cls.env['openerp.enterprise.database']

    def setUp(self):
        super().setUp()
        self.cr.precommit.clear()

    def _register(self, domain, contact_type, unlink=True, **extra_params):

        def _fake_domain_info_error(*args, **kwargs):
            raise Exception('Some API error')

        if contact_type != ContactType.person:
            extra_params['orgname'] = 'odoo'

        with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.info', new=_fake_domain_info_error, create=True):
            oe_domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id) \
                ._register_domain(self.db.id, domain, contact_type, 'foo', 'bar', 'foo@bar.baz', '0032490000000', 'BE', 'addr', 'city', '1000', **extra_params)

        self.assertEqual(oe_domain.fqdn, domain)
        self.assertEqual(oe_domain.user_id, self.owner)
        self.assertEqual(oe_domain.expiration_date, fields.Date.add(fields.Date.today(), years=1))
        self.assertEqual(oe_domain.subscription_id, self.sub)

        if unlink:
            oe_domain.unlink()

    def _make_one_app_free(self):
        version_id = self.env['openerp.enterprise.module.version']._get_latest_version()
        website_id = version_id.module_ids.filtered(lambda x: x.technical_name == 'website').id
        return self.env['openerp.enterprise.database'].create({
            'db_uuid': str(uuid.uuid4()),
            'db_name': 'mycompany',
            'hosting': 'saas',
            'last_ping': fields.Date.subtract(fields.Date.today(), days=29),
            'registered_user_ids': [(0, 0, {'user_id': self.owner.id, 'is_admin': True})],
            'ping_ids': [(0, 0, {
                'ping_time': fields.Date.subtract(fields.Date.today(), days=29),
                'ip': '10.53.0.82',  # 127.0.0.1
                'db_name': 'mycompany',
                'version': 14.0,
                'nbr_users': 1,
                'language': 'fr_BE',
                'extra_apps': False,
            })],
            'allow_free_domain': True,
            'app_ids': [(0, 0, {
                'module_id': website_id,
                'installed': True,
            })],
        })

    def _make_db(self, subscription=None):
        version_id = self.env['openerp.enterprise.module.version']._get_latest_version()
        website_id = version_id.module_ids.filtered(lambda x: x.technical_name == 'website').id
        return self.env['openerp.enterprise.database'].create({
            'db_uuid': str(uuid.uuid4()),
            'db_name': 'mycompany',
            'hosting': 'saas',
            'last_ping': fields.Date.subtract(fields.Date.today(), days=29),
            'registered_user_ids': [(0, 0, {'user_id': self.owner.id, 'is_admin': True})],
            'ping_ids': [(0, 0, {
                'ping_time': fields.Date.subtract(fields.Date.today(), days=29),
                'ip': '10.53.0.82',  # 127.0.0.1
                'db_name': 'mycompany',
                'version': 14.0,
                'nbr_users': 1,
                'language': 'fr_BE',
                'extra_apps': True,
            })],
            'subscription_id': subscription.id if subscription else None,
            'app_ids': [(0, 0, {
                'module_id': website_id,
                'installed': True,
            })],
        })

    def _make_unpaid_database(self):
        return self._make_db()

    def _make_paid_database(self, subscription):
        return self._make_db(subscription)

    def _subscribe(self):
        self.sub = self.env['sale.order'].create({
            'partner_id': self.owner.partner_id.id,
            'enterprise_final_customer_id': self.owner.partner_id.id,
            'sale_order_template_id': self.env.ref('openerp_enterprise.standard_template_odoo_yearly').id,
        })
        self.sub._onchange_sale_order_template_id()
        self.sub.action_confirm()
        self.db = self._make_paid_database(self.sub)

@tagged('oe_public_suffix_list')
class TestOdooPublicSuffixList(TransactionCase):

    def test_public_suffix_list(self):
        def checkPublicSuffix(domain, expected):
            self.assertEqual(PSL.sld(domain), expected)

        # The following tests (adapted to python) are taken from https://raw.githubusercontent.com/publicsuffix/list/master/tests/test_psl.txt

        # None input.
        checkPublicSuffix(None, None)
        # Mixed case.
        checkPublicSuffix('COM', None)
        checkPublicSuffix('example.COM', 'example.com')
        checkPublicSuffix('WwW.example.COM', 'example.com')
        # Leading dot.
        checkPublicSuffix('.com', None)
        checkPublicSuffix('.example', None)
        checkPublicSuffix('.example.com', None)
        checkPublicSuffix('.example.example', None)
        # Unlisted TLD.
        checkPublicSuffix('example', None)
        checkPublicSuffix('example.example', 'example.example')
        checkPublicSuffix('b.example.example', 'example.example')
        checkPublicSuffix('a.b.example.example', 'example.example')
        # TODO(jde): block list private domains.
        # Don't see a use case for now. Let's see if we need it in the future.
        # https://www.iana.org/assignments/special-use-domain-names/special-use-domain-names.xhtml
        # https://www.rfc-editor.org/rfc/rfc6762#appendix-G
        # Listed, but non-Internet, TLD.
        # checkPublicSuffix('local', None)
        # checkPublicSuffix('example.local', None)
        # checkPublicSuffix('b.example.local', None)
        # checkPublicSuffix('a.b.example.local', None)
        # TLD with only 1 rule.
        checkPublicSuffix('biz', None)
        checkPublicSuffix('domain.biz', 'domain.biz')
        checkPublicSuffix('b.domain.biz', 'domain.biz')
        checkPublicSuffix('a.b.domain.biz', 'domain.biz')
        # TLD with some 2-level rules.
        checkPublicSuffix('com', None)
        checkPublicSuffix('example.com', 'example.com')
        checkPublicSuffix('b.example.com', 'example.com')
        checkPublicSuffix('a.b.example.com', 'example.com')
        checkPublicSuffix('uk.com', None)
        checkPublicSuffix('example.uk.com', 'example.uk.com')
        checkPublicSuffix('b.example.uk.com', 'example.uk.com')
        checkPublicSuffix('a.b.example.uk.com', 'example.uk.com')
        checkPublicSuffix('test.ac', 'test.ac')
        # TLD with only 1 (wildcard) rule.
        checkPublicSuffix('mm', None)
        checkPublicSuffix('c.mm', None)
        checkPublicSuffix('b.c.mm', 'b.c.mm')
        checkPublicSuffix('a.b.c.mm', 'b.c.mm')
        # More complex TLD.
        checkPublicSuffix('jp', None)
        checkPublicSuffix('test.jp', 'test.jp')
        checkPublicSuffix('www.test.jp', 'test.jp')
        checkPublicSuffix('ac.jp', None)
        checkPublicSuffix('test.ac.jp', 'test.ac.jp')
        checkPublicSuffix('www.test.ac.jp', 'test.ac.jp')
        checkPublicSuffix('kyoto.jp', None)
        checkPublicSuffix('test.kyoto.jp', 'test.kyoto.jp')
        checkPublicSuffix('ide.kyoto.jp', None)
        checkPublicSuffix('b.ide.kyoto.jp', 'b.ide.kyoto.jp')
        checkPublicSuffix('a.b.ide.kyoto.jp', 'b.ide.kyoto.jp')
        checkPublicSuffix('c.kobe.jp', None)
        checkPublicSuffix('b.c.kobe.jp', 'b.c.kobe.jp')
        checkPublicSuffix('a.b.c.kobe.jp', 'b.c.kobe.jp')
        checkPublicSuffix('city.kobe.jp', 'city.kobe.jp')
        checkPublicSuffix('www.city.kobe.jp', 'city.kobe.jp')
        # TLD with a wildcard rule and exceptions.
        checkPublicSuffix('ck', None)
        checkPublicSuffix('test.ck', None)
        checkPublicSuffix('b.test.ck', 'b.test.ck')
        checkPublicSuffix('a.b.test.ck', 'b.test.ck')
        checkPublicSuffix('www.ck', 'www.ck')
        checkPublicSuffix('www.www.ck', 'www.ck')
        # US K12.
        checkPublicSuffix('us', None)
        checkPublicSuffix('test.us', 'test.us')
        checkPublicSuffix('www.test.us', 'test.us')
        checkPublicSuffix('ak.us', None)
        checkPublicSuffix('test.ak.us', 'test.ak.us')
        checkPublicSuffix('www.test.ak.us', 'test.ak.us')
        checkPublicSuffix('k12.ak.us', None)
        checkPublicSuffix('test.k12.ak.us', 'test.k12.ak.us')
        checkPublicSuffix('www.test.k12.ak.us', 'test.k12.ak.us')
        # IDN labels.
        checkPublicSuffix('食狮.com.cn', '食狮.com.cn')
        checkPublicSuffix('食狮.公司.cn', '食狮.公司.cn')
        checkPublicSuffix('www.食狮.公司.cn', '食狮.公司.cn')
        checkPublicSuffix('shishi.公司.cn', 'shishi.公司.cn')
        checkPublicSuffix('公司.cn', None)
        checkPublicSuffix('食狮.中国', '食狮.中国')
        checkPublicSuffix('www.食狮.中国', '食狮.中国')
        checkPublicSuffix('shishi.中国', 'shishi.中国')
        checkPublicSuffix('中国', None)
        # Same as above, but punycoded.
        checkPublicSuffix('xn--85x722f.com.cn', 'xn--85x722f.com.cn')
        checkPublicSuffix('xn--85x722f.xn--55qx5d.cn', 'xn--85x722f.xn--55qx5d.cn')
        checkPublicSuffix('www.xn--85x722f.xn--55qx5d.cn', 'xn--85x722f.xn--55qx5d.cn')
        checkPublicSuffix('shishi.xn--55qx5d.cn', 'shishi.xn--55qx5d.cn')
        checkPublicSuffix('xn--55qx5d.cn', None)
        checkPublicSuffix('xn--85x722f.xn--fiqs8s', 'xn--85x722f.xn--fiqs8s')
        checkPublicSuffix('www.xn--85x722f.xn--fiqs8s', 'xn--85x722f.xn--fiqs8s')
        checkPublicSuffix('shishi.xn--fiqs8s', 'shishi.xn--fiqs8s')
        checkPublicSuffix('xn--fiqs8s', None)

    def test_extract(self):
        self.assertEqual(PSL.extract('com'), None)
        self.assertEqual(PSL.extract('example.com'), ('', 'example', 'com'))
        self.assertEqual(PSL.extract('www.example.com'), ('www', 'example', 'com'))
        self.assertEqual(PSL.extract('a.b.www.example.com'), ('a.b.www', 'example', 'com'))
        self.assertEqual(PSL.extract('example.uk.com'), ('', 'example', 'uk.com'))
        self.assertEqual(PSL.extract('b.example.uk.com'), ('b', 'example', 'uk.com'))
        self.assertEqual(PSL.extract('a.b.example.uk.com'), ('a.b', 'example', 'uk.com'))
        self.assertEqual(PSL.extract('a.b.www.example.uk.com'), ('a.b.www', 'example', 'uk.com'))
        self.assertEqual(PSL.extract('test.kyoto.jp'), ('', 'test', 'kyoto.jp'))
        self.assertEqual(PSL.extract('c.kobe.jp'), None)
        self.assertEqual(PSL.extract('b.c.kobe.jp'), ('', 'b', 'c.kobe.jp'))
        self.assertEqual(PSL.extract('a.b.c.kobe.jp'), ('a', 'b', 'c.kobe.jp'))
        self.assertEqual(PSL.extract('city.kobe.jp'), ('', 'city', 'kobe.jp'))
        self.assertEqual(PSL.extract('www.city.kobe.jp'), ('www', 'city', 'kobe.jp'))
        self.assertEqual(PSL.extract('b.test.ck'), ('', 'b', 'test.ck'))
        self.assertEqual(PSL.extract('a.b.test.ck'), ('a', 'b', 'test.ck'))
        self.assertEqual(PSL.extract('www.ck'), ('', 'www', 'ck'))
        self.assertEqual(PSL.extract('www.www.ck'), ('www', 'www', 'ck'))
        self.assertEqual(PSL.extract('www.食狮.公司.cn'), ('www', '食狮', '公司.cn'))
        self.assertEqual(PSL.extract('www.xn--85x722f.xn--fiqs8s'), ('www', 'xn--85x722f', 'xn--fiqs8s'))


@tagged('oe_dns')
class TestOdooDNS(EnterpriseDatabaseCommon, TestDNSSetup, MockEmail):
    # These tests mock the registrar's API responses and test how things behave on Odoo's side.

    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._status')
    def test_search_domain(self, status_mock):
        status_mock.return_value = 'available'

        def _status(*args, **kwargs):
            if args[0] in ['odoo.com', 'www.odoo.com']:
                return 'unavailable'
            if args[0] == 'odoo.online':
                return 'error'
            return 'available'

        def filter_suggestions(res):
            return {key: value for key, value in res.items() if value != 'pending'}

        status_mock.side_effect = _status

        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        # User input + suggestions.
        res = Domain._search_domain('example.co.uk')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'example.co.uk': 'available',
            'example.com': 'available',
            'example.net': 'available',
            'example.org': 'available',
        })

        # User input was going to be suggested, don't suggest it again.
        res = Domain._search_domain('example.com')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'example.com': 'available',
            'example.net': 'available',
            'example.org': 'available',
        })

        # User input goes first.
        res = Domain._search_domain('example.org')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'example.org': 'available',
            'example.com': 'available',
            'example.net': 'available',
        })

        # Show exact domain name entered by user, even if not available.
        res = Domain._search_domain('odoo.com')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'odoo.com': 'unavailable',
            'odoo.net': 'available',
            'odoo.org': 'available',
        })

        # Don't show unavailable domains if no TLD is given.
        res = Domain._search_domain('odoo')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'odoo.net': 'available',
            'odoo.org': 'available',
        })

        # TLD not supported: status=error.
        res = Domain._search_domain('odoo.online')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'odoo.online': 'error',
            'odoo.net': 'available',
            'odoo.org': 'available',
        })

        # Remove sudomain before checking status but keep user input untouched.
        res = Domain._search_domain('www.odoo.com')
        res = filter_suggestions(res)
        self.assertEqual(res, {
            'www.odoo.com': 'unavailable',
            'odoo.net': 'available',
            'odoo.org': 'available',
        })

    def test_domain_availability(self):
        assert TEST, 'aborting...'

        domain_type = ''
        price = 15.0

        def _fake_response(*args, **kwargs):
            nonlocal domain_type
            nonlocal price
            return {
                'products': [
                    {
                        'prices': [
                            {
                                'type': domain_type,
                                'price_after_taxes': price,
                                'min_duration': 2 if args[1] == 'example.gr' else 1
                            },
                        ],
                        'status': 'available',
                    },
                ],
            }

        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status', new=_fake_response, create=True):
            # Ensure TLD is set.
            with self.assertRaises(ValueError) as ex:
                Domain._status('example')
            self.assertEqual(ex.exception.args[0], 'A top-level domain is required (e.g. example.com).')

            # Ensure IDN is rejected.
            with self.assertRaises(ValueError) as ex:
                Domain._status('bücher.com')
            self.assertEqual(ex.exception.args[0], 'IDN not supported.')

            # Ensure TLD is part of our offering.
            with self.assertRaises(ValidationError) as ex:
                Domain._status('example.website')
            self.assertTrue(ex.exception.args[0], 'TLD not supported.')

            # Ensure premium domains are rejected.
            domain_type = 'premium'
            with self.assertRaises(Exception) as ex:
                Domain._status('bizcom.net')
            self.assertEqual(ex.exception.args[0], 'Premium domain names are not supported.')
            domain_type = ''

            # Ensure price is within our treshold.
            price = 999.99
            with self.assertRaises(Exception) as ex:
                Domain._status('example.com')
            self.assertEqual(ex.exception.args[0], 'This domain is not covered by our free plan.')

            # Ensure min_duration is checked.
            # 1 year price = 16 but requires a 2 year commitment -> total = 32. Should be rejected.
            price = 16
            with self.assertRaises(Exception) as ex:
                Domain._status('example.gr')
            self.assertEqual(ex.exception.args[0], 'This domain is not covered by our free plan.')

            price = 15.0
            # Ensure the domain status is correctly returned.
            self.assertEqual(Domain._status('example.com'), 'available')

    @patch('requests.request')
    def test_gandi_status_http_request(self, requests_mock):
        fqdn = f"{uuid.uuid4()}.com"

        GANDI.status(fqdn)

        requests_mock.assert_called_once()
        args = requests_mock.call_args[0]
        kwargs = requests_mock.call_args[1]
        self.assertEqual(args, ('GET', 'https://api.gandi.net/v5/domain/check'))
        self.assertEqual(kwargs['params']['currency'], 'EUR', 'Bad currency')
        self.assertEqual(kwargs['params']['name'], fqdn, 'Bad fqdn')
        self.assertEqual(kwargs.get('json'), None, 'Bad payload')

    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._status')
    @patch('requests.request')
    def test_gandi_register_domain_http_request(self, requests_mock, status_mock):
        """Validates payload sent to Gandi."""
        requests_mock.return_value.json.return_value = {'status': 'success'}
        status_mock.return_value = 'available'

        # Expected data.
        fqdn = f"{uuid.uuid4()}.com"
        owner_data = {
            'type': ContactType.person.value,
            'given': self.input_data['fname'],
            'family': self.input_data['lname'],
            'email': self.input_data['email'],
            'phone': self.input_data['phone'],
            'country': self.input_data['country'],
            'streetaddr': self.input_data['address'],
            'city': self.input_data['city'],
            'zip': self.input_data['zip'],
        }

        # Trigger http call (mocked).
        GANDI.register_domain(fqdn, owner_data, dry_run=True)

        # Verify endpoint and payload.
        requests_mock.assert_called_once()
        args = requests_mock.call_args[0]
        kwargs = requests_mock.call_args[1]
        self.assertEqual(args, ('POST', 'https://api.gandi.net/v5/domain/domains'))
        self.assertEqual(kwargs.get('json'), {
            'fqdn': fqdn,
            'duration': 1,
            'owner': {
                'type': 0,
                'given': self.input_data['fname'],
                'family': self.input_data['lname'],
                'email': self.input_data['email'],
                'phone': self.input_data['phone'],
                'country': self.input_data['country'],
                'streetaddr': self.input_data['address'],
                'city': self.input_data['city'],
                'zip': self.input_data['zip'],
                'data_obfuscated': True,
                'mail_obfuscated': True,
                'extra_parameters': {},
            },
            'admin': self.odoo_sa,
            'bill': self.odoo_sa,
            'tech': self.odoo_sa,
            'currency': 'EUR',
        }, 'Bad payload')

    @patch('requests.request')
    def test_gandi_info_http_request(self, requests_mock):
        fqdn = f"{uuid.uuid4()}.com"

        GANDI.info(fqdn)

        requests_mock.assert_called_once()
        args = requests_mock.call_args[0]
        kwargs = requests_mock.call_args[1]
        self.assertEqual(args, ('GET', f'https://api.gandi.net/v5/domain/domains/{fqdn}'))
        self.assertEqual(kwargs.get('json'), None, 'Bad payload')

    @patch('requests.request')
    def test_gandi_create_web_redirection_http_request(self, requests_mock):
        fqdn = f"{uuid.uuid4()}.com"

        GANDI.create_web_redirection(fqdn, fqdn, f'https://www.{fqdn}', 'http301')

        requests_mock.assert_called_once()
        args = requests_mock.call_args[0]
        kwargs = requests_mock.call_args[1]
        self.assertEqual(args, ('POST', f'https://api.gandi.net/v5/domain/domains/{fqdn}/webredirs'))
        self.assertEqual(kwargs.get('json'), {
            'host': fqdn,
            'type': 'http301',
            'url': f'https://www.{fqdn}',
            'protocol': 'https',
            'override': True,
        }, 'Bad payload')

    @patch('requests.request')
    def test_gandi_delete_records_http_request(self, requests_mock):
        fqdn = f"{uuid.uuid4()}.com"

        GANDI.delete_records(fqdn)

        requests_mock.assert_called_once()
        args = requests_mock.call_args[0]
        kwargs = requests_mock.call_args[1]
        self.assertEqual(args, ('DELETE', f'https://api.gandi.net/v5/livedns/domains/{fqdn}/records'))
        self.assertEqual(kwargs.get('json'), None, 'Bad payload')

    @patch('requests.request')
    def test_gandi_put_record_http_request(self, requests_mock):
        fqdn = f"{uuid.uuid4()}.com"

        GANDI.put_record(fqdn, 'CNAME', 'www', [f'mydatabase.{SAAS_BASE_URL}.'], ttl=600)

        requests_mock.assert_called_once()
        args = requests_mock.call_args[0]
        kwargs = requests_mock.call_args[1]
        self.assertEqual(args, ('PUT', f'https://api.gandi.net/v5/livedns/domains/{fqdn}/records/www/CNAME'))
        self.assertEqual(kwargs.get('json'), {
            'rrset_values': [f'mydatabase.{SAAS_BASE_URL}.'],
            'rrset_ttl': 600,
        }, 'Bad payload')

    @patch('requests.post')
    def test_saas_worker_http_request(self, post_mock):
        """Validates payload sent to the trial server on which the database is hosted."""
        fqdn = f"{uuid.uuid4()}.com"
        self._subscribe()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': fqdn,
            'user_id': self.owner.id,
            'expiration_date': fields.Date.to_date('2021-08-03'),
            'delete_date': fields.Date.to_date('2021-10-22'),
            'database_id': self.db.id,
            'synced': True,
        })

        domain._route_hostname_to_db(f'www.{fqdn}')

        post_mock.assert_called_once()
        args = post_mock.call_args[0]
        kwargs = post_mock.call_args[1]
        self.assertEqual(args, (f'http://{self.db.db_name}.{SAAS_BASE_URL}/saas_worker/domain-names/trusted-add/',))
        self.assertEqual(kwargs.get('data'), {
            'db_name': self.db.db_name,
            'hostname': f'www.{fqdn}',
        })

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.info')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.register_domain')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status')
    @freeze_time('2023-01-01')
    def test_odoo_register_domain_happy_path(self, status_mock, register_mock, info_mock):
        """Validates odoo's behavior when everything goes right.
        Responses to requests made to Gandi are mocked."""
        # Mocks:
        status_mock.return_value = {
            'products': [
                {
                    'prices': [{'type': '', 'price_after_taxes': 10.0}],
                    'status': 'available',
                },
            ],
        }
        register_mock.return_value = {'status': 'success'}
        info_mock.return_value = {
            'contacts': {
                'owner': {
                    'reachability': 'done',
                },
            },
            'dates': {
                'hold_begins_at': '2024-01-01T00:00:Z',
                'pending_delete_ends_at': '2024-03-21T00:00:00Z',  # 2024-01-01 + 80 days
            },
        }

        # Being subscribed is a prerequisite:
        self._subscribe()

        # Input data:
        fqdn = f"{uuid.uuid4()}.com"

        # Trigger domain registration:
        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)
        oe_domain = Domain._register_domain(
            self.db.id,
            fqdn,
            ContactType.person.value,
            self.input_data['fname'],
            self.input_data['lname'],
            self.input_data['email'],
            self.input_data['phone'],
            self.input_data['country'],
            self.input_data['address'],
            self.input_data['city'],
            zip=self.input_data['zip'],
        )

        # Verify what is inserted in the database (openerp_enterprise_domain table):
        self.assertEqual(oe_domain.fqdn, fqdn)
        self.assertEqual(oe_domain.user_id, self.owner)
        expiration_date = fields.Date.add(fields.Date.today(), months=11)
        self.assertEqual(oe_domain.expiration_date, expiration_date)
        self.assertEqual(oe_domain.delete_date, oe_domain.expiration_date)
        self.assertEqual(oe_domain.database_id, self.db)
        self.assertEqual(oe_domain.subscription_id, self.db.subscription_id)
        self.assertEqual(oe_domain.subscription_id, self.sub)
        self.assertEqual(oe_domain.synced, False)

        oe_domain._sync()
        self.assertEqual(oe_domain.expiration_date, fields.Date.to_date('2024-01-01'))
        self.assertEqual(oe_domain.delete_date, fields.Date.to_date('2024-03-21'))
        self.assertEqual(oe_domain.synced, True)

        # Verify Gandi has been called with expected inputs:
        expected_owner = {
            'type': ContactType.person.value,
            'given': self.input_data['fname'],
            'family': self.input_data['lname'],
            'email': self.input_data['email'],
            'phone': self.input_data['phone'],
            'country': self.input_data['country'],
            'streetaddr': self.input_data['address'],
            'city': self.input_data['city'],
            'zip': self.input_data['zip'],
        }
        register_mock.assert_called_once_with(fqdn, expected_owner, dry_run=True)
        info_mock.assert_called_once_with(fqdn)

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.register_domain')
    def test_odoo_register_domain_nok_domain_not_allowed(self, register_mock):
        fqdn = f"{uuid.uuid4()}.com"
        db = self._make_one_app_free()

        self.env['openerp.enterprise.trial.rules'].create({
            'name': 'SEO: casino poker',
            'type': 'name',
            'rule': '(slot|casino|poker|lottery)',
        })

        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        with self.assertRaises(ValidationError) as ex:
            Domain._register_domain(
                db.id,
                f"casino-{fqdn}",
                ContactType.person.value,
                self.input_data['fname'],
                self.input_data['lname'],
                self.input_data['email'],
                self.input_data['phone'],
                self.input_data['country'],
                self.input_data['address'],
                self.input_data['city'],
                zip=self.input_data['zip'],
            )

        self.assertEqual(ex.exception.args[0], 'Domain name not allowed.')
        register_mock.assert_not_called()

    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._status')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.register_domain')
    def test_odoo_register_domain_nok_database_required(self, register_mock, status_mock):
        status_mock.return_value = 'available'

        fqdn = f"{uuid.uuid4()}.com"
        db = self._make_one_app_free()
        # Remove test user from registered users. Since it's required to be a
        # registered user in order to book a domain name, the test should fail.
        db.write({'registered_user_ids': [(5, 0, 0)]})
        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        with self.assertRaises(ValidationError) as ex:
            Domain._register_domain(
                db.id,
                fqdn,
                ContactType.person.value,
                self.input_data['fname'],
                self.input_data['lname'],
                self.input_data['email'],
                self.input_data['phone'],
                self.input_data['country'],
                self.input_data['address'],
                self.input_data['city'],
                zip=self.input_data['zip'],
            )

        self.assertEqual(ex.exception.args[0], 'Invalid database.')
        register_mock.assert_not_called()

    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._status')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.register_domain')
    def test_odoo_register_domain_nok_one_app_free_support_clearance_required(self, register_mock, status_mock):
        status_mock.return_value = 'available'

        fqdn = f"{uuid.uuid4()}.com"
        db = self._make_one_app_free()
        db.allow_free_domain = False

        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        request = Domain._register_domain(
            db.id,
            fqdn,
            ContactType.person.value,
            self.input_data['fname'],
            self.input_data['lname'],
            self.input_data['email'],
            self.input_data['phone'],
            self.input_data['country'],
            self.input_data['address'],
            self.input_data['city'],
            zip=self.input_data['zip'],
        )

        self.assertTrue(request.exists(), 'A request should be created')
        self.assertEqual(request.fqdn, fqdn)
        self.assertEqual(request.user_id, self.owner)
        self.assertEqual(request.database_id, db)
        self.assertEqual(request.state, 'pending')
        self.assertEqual(json.loads(request.payload), {
            'type': ContactType.person.value,
            'given': self.input_data['fname'],
            'family': self.input_data['lname'],
            'email': self.input_data['email'],
            'phone': self.input_data['phone'],
            'country': self.input_data['country'],
            'streetaddr': self.input_data['address'],
            'city': self.input_data['city'],
            'zip': self.input_data['zip'],
        })
        register_mock.assert_not_called()

        # Grant free domain name.
        request.action_grant()
        register_mock.assert_called()
        self.assertTrue(db.allow_free_domain)
        self.assertEqual(request.state, 'granted')

    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._status')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.register_domain')
    def test_odoo_register_domain_nok_subscription_required(self, register_mock, status_mock):
        """A valid enterprise subscription is required to register a free domain name."""
        status_mock.return_value = 'available'

        fqdn = f"{uuid.uuid4()}.com"
        db = self._make_unpaid_database()
        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        with self.assertRaises(ValidationError) as ex:
            Domain._register_domain(
                db.id,
                fqdn,
                ContactType.person.value,
                self.input_data['fname'],
                self.input_data['lname'],
                self.input_data['email'],
                self.input_data['phone'],
                self.input_data['country'],
                self.input_data['address'],
                self.input_data['city'],
                zip=self.input_data['zip'],
            )

        self.assertEqual(ex.exception.args[0], 'Your Odoo Subscription is not valid. Go to your subscription or contact us.')
        register_mock.assert_not_called()

    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._status')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.register_domain')
    def test_odoo_register_domain_nok_subscription_already_used(self, register_mock, status_mock):
        """A subscription can only be used once."""
        status_mock.return_value = 'available'

        self._subscribe()
        self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'user_id': self.owner.id,
            'expiration_date': fields.Date.to_date('2021-08-03'),
            'delete_date': fields.Date.to_date('2021-10-22'),
            'database_id': self.db.id,
            'synced': True,
        })
        Domain = self.env['openerp.enterprise.domain'].with_user(self.owner.id)

        another_db = self._make_paid_database(self.sub)
        with self.assertRaises(ValidationError) as ex:
            Domain._register_domain(
                another_db.id,
                f'{uuid.uuid4()}.com',
                ContactType.person.value,
                self.input_data['fname'],
                self.input_data['lname'],
                self.input_data['email'],
                self.input_data['phone'],
                self.input_data['country'],
                self.input_data['address'],
                self.input_data['city'],
                zip=self.input_data['zip'],
            )

        self.assertEqual(ex.exception.args[0], f'Subscription {self.sub.client_order_ref} is already in use.')
        register_mock.assert_not_called()

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status')
    def test_renew_price(self, status_mock):
        self._subscribe()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-01-14'),
            'delete_date': fields.Date.to_date('2021-04-04'),
            'user_id': self.owner.id,
            'database_id': self.db.id,
            'synced': True,
            'email_confirmed': True,
        })

        status_mock.return_value = {
            'products': [{'prices': [{'price_before_taxes': 10.0}]}],
        }
        self.assertEqual(domain._renew_price(), 10.0)

        status_mock.return_value = {
            'products': [{'prices': [{
                'price_before_taxes': 10.0,
                'min_duration': 2
        }]}]}
        self.assertEqual(domain._renew_price(), 20.0)

    @freeze_time('2021-01-01')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status')
    def test_quote_renewal(self, status_mock):
        status_mock.return_value = {
            'products': [{'prices': [{'price_before_taxes': 10.0}]}],
        }

        self.owner.country_id = self.env.ref('base.be')
        self.env['product.pricelist'].create({
            'name': 'EUROPE EUR',
            'country_group_ids': [fields.Command.link(self.env.ref('base.europe').id)],
            'sequence': 1,
            'currency_id': self.env.ref('base.EUR').id,
        })

        self._subscribe()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-01-30'),
            'delete_date': fields.Date.to_date('2021-04-04'),
            'user_id': self.owner.id,
            'database_id': self.db.id,
            'synced': True,
            'email_confirmed': True,
        })
        self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)

        company = self.env.ref('base.main_company')
        pl = domain.user_id.partner_id.property_product_pricelist
        expected_price = pl.currency_id.round(
            self.env.ref('base.EUR')._convert(10.0, pl.currency_id, company, fields.Date.today())
            + domain.renew_order_id.amount_tax
        )

        self.assertTrue(domain.renew_order_id)
        self.assertEqual(domain.renew_order_id.state, 'sent')
        self.assertEqual(domain.renew_order_id.domain_id, domain)
        self.assertEqual(domain.renew_order_id.amount_total, expected_price)
        self.assertEqual(domain.renew_order_id.currency_id, pl.currency_id)
        self.assertEqual(domain.renew_order_id.currency_id, self.env.ref('base.EUR'))
        self.assertEqual(domain.renew_order_id.payment_term_id, self.env.ref('openerp_enterprise.payment_term_immediate'))
        self.assertEqual(domain.renew_order_id.validity_date, domain.expiration_date)
        self.assertEqual(domain.renew_order_count, 1)
        self.assertEqual(domain.to_renew, False)

        self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)
        self.assertEqual(domain.renew_order_count, 1, "Existing quotation: don't create a new one")
        self.assertEqual(domain.to_renew, False)

    @freeze_time('2021-01-01')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status')
    def test_regress_quote_renewal_currency(self, status_mock):
        """
        Since Odoo S.A. only accepts payments in EUR or USD through Ogone,
        force USD when the customer currency is not one of these two.

        Here we first make sure the chosen currency for the partner is MXN.
        We then check that the sale order is using USD.
        """
        status_mock.return_value = {
            'products': [{'prices': [{'price_before_taxes': 10.0}]}],
        }

        self.owner.country_id = self.env.ref('base.mx')
        latam = self.env['res.country.group'].create({
            'name': 'LATAM',
            'country_ids': [fields.Command.link(self.env.ref('base.mx').id)],
        })
        mx_pl = self.env['product.pricelist'].create({
            'name': 'MEXICO',
            'country_group_ids': [fields.Command.link(latam.id)],
            'sequence': 1,
            'currency_id': self.env.ref('base.MXN').id,
        })
        self.assertEqual(self.owner.partner_id.property_product_pricelist, mx_pl)

        self._subscribe()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-01-30'),
            'delete_date': fields.Date.to_date('2021-04-04'),
            'user_id': self.owner.id,
            'database_id': self.db.id,
            'synced': True,
            'email_confirmed': True,
        })
        self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)

        self.assertEqual(self.owner.partner_id.property_product_pricelist.currency_id, self.env.ref('base.MXN'))
        self.assertEqual(domain.renew_order_id.currency_id, self.env.ref('base.USD'))

    @freeze_time('2021-01-01')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.renew_domain')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status')
    def test_payment_and_renewal_flow(self, status_mock, renew_mock):
        status_mock.return_value = {
            'products': [{'prices': [{'price_before_taxes': 10.0}]}],
        }
        renew_mock.return_value = True

        self._subscribe()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-01-30'),
            'delete_date': fields.Date.to_date('2021-04-20'),
            'user_id': self.owner.id,
            'database_id': self.db.id,
            'synced': True,
            'email_confirmed': True,
        })

        # Quote the renewal.
        self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)
        self.assertTrue(domain.renew_order_id, 'sale order not created')
        self.assertEqual(domain.to_renew, False)

        # Confirm and invoice the sale order.
        domain.renew_order_id.action_confirm()
        self.assertEqual(domain.renew_order_id.state, 'sale')
        self.assertEqual(domain.to_renew, False)
        invoice = domain.renew_order_id._create_invoices()
        self.assertEqual(len(domain.renew_order_id.invoice_ids), 1)
        invoice.action_post()
        self.assertEqual(domain.renew_order_id.invoice_status, 'invoiced')
        self.assertEqual(domain.renew_order_id.invoice_count, 1)
        self.assertEqual(domain.to_renew, False)

        # Pay the invoice.
        journal = self.env['account.journal'].search([
            ('type', '=', 'cash'),
            ('company_id', '=', domain.renew_order_id.company_id.id)
        ], limit=1)
        ctx = {'active_model': 'account.move', 'active_ids': [invoice.id]}
        payment_register = self.env['account.payment.register'].with_context(ctx).create({
            'journal_id': journal.id,
        })
        payment_register._create_payments()
        in_payment_state = invoice._get_invoice_in_payment_state()
        self.assertEqual(invoice.payment_state, in_payment_state)
        self.assertEqual(domain.renew_order_id.state, 'sale')
        self.assertEqual(domain.to_renew, True)

        # Run the cron job again to trigger renewal.
        self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)
        renew_mock.assert_called_once_with(domain.fqdn, dry_run=True)
        self.assertEqual(domain.expiration_date, fields.Date.to_date('2021-12-30'))
        self.assertEqual(domain.delete_date, fields.Date.to_date('2021-12-30'), "must equal expiration date")
        self.assertFalse(domain.renew_order_id)
        self.assertEqual(len(domain.renew_order_ids), 1)
        self.assertEqual(domain.renew_order_count, 1)
        self.assertFalse(domain.to_renew)

    def test_renew(self):
        assert TEST, 'aborting...'

        self._subscribe()

        def renew_day(*args, **kwargs):
            return fields.Date.to_date('2021-01-01')

        renew_call_count = 0

        def _fake_renew_domain(*args, **kwargs):
            nonlocal renew_call_count
            renew_call_count += 1
            return True

        def _fake_domain_info(*args, **kwargs):
            return {
                'contacts': {
                    'owner': {
                        'reachability': 'done',
                    },
                },
                'dates': {
                    'hold_begins_at': '2022-01-30T13:03:42Z',
                    'pending_delete_ends_at': '2022-04-20T13:03:42Z',  # 2022-01-30 + 80 days
                },
            }

        def _set_to_renew(d):
            product = self.env.ref('openerp_enterprise.product_domain_name_renewal')
            so = self.env['sale.order'].create({
                'partner_id': d.user_id.partner_id.id,
                'domain_id': d.id,
                'order_line': [fields.Command.create({
                    'product_id': product.id,
                    'price_unit': 10.0,
                    'tax_id': False,  # VAT is not charged by registrar (intra-community)
                    'product_uom_qty': 1,
                })],
                'require_payment' : True,
                'validity_date': d.expiration_date,
            })
            so.action_confirm()
            d.renew_order_id = so.id
            d.to_renew = True

        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-01-30'),
            'delete_date': fields.Date.to_date('2021-04-20'),  # 2021-01-30 + 80 days
            'user_id': self.owner.id,
            'database_id': self.db.id,
            'synced': True,
            'email_confirmed': True,
        })
        _set_to_renew(domain)

        with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.renew_domain', new=_fake_renew_domain, create=True):
            with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.info', new=_fake_domain_info, create=True):
                with patch('odoo.fields.Date.today', renew_day):
                    # Test the expiration_date is extended for a year, 14 days beforehand.
                    with self.subTest('OK: extend 1 year'):
                        self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)
                        self.assertEqual(domain.expiration_date, fields.Date.to_date('2021-12-30'))
                        self.assertEqual(domain.delete_date, fields.Date.to_date('2021-12-30'), "must equal expiration date")
                        self.assertEqual(domain.synced, False)
                        self.assertEqual(renew_call_count, 1)
                        self.assertFalse(domain.renew_order_id)
                        self.assertFalse(domain.to_renew)

                    with self.subTest('Post-renew data synchronization'):
                        self.env['openerp.enterprise.domain']._cron_sync_data_with_registrar(auto_commit=False)
                        self.assertEqual(domain.synced, True)
                        self.assertEqual(domain.expiration_date, fields.Date.to_date('2022-01-30'))
                        self.assertEqual(domain.delete_date, fields.Date.to_date('2022-04-20'))  # 2022-01-30 + 80 days

    def test_sync_data_with_registrar(self):
        assert TEST, 'aborting...'

        def _fake_domain_info(*args, **kwargs):
            return {
                'dates': {
                    'hold_begins_at': '2021-08-03T13:03:42Z',
                    'pending_delete_ends_at': '2021-10-22T13:03:42Z',
                },
                'contacts': {
                    'owner': {
                        'reachability': 'done',
                    },
                },
            }

        db1 = self._make_one_app_free()
        d1 = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-08-02'),
            'delete_date': fields.Date.to_date('2021-10-20'),
            'user_id': self.owner.id,
            'database_id': db1.id,
        })

        db2 = self._make_one_app_free()
        d2 = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2021-08-02'),
            'delete_date': fields.Date.to_date('2021-10-20'),
            'user_id': self.owner.id,
            'database_id': db2.id,
            'synced': True,
            'email_confirmed': False,
        })

        with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.info', new=_fake_domain_info, create=True):
            self.env['openerp.enterprise.domain']._cron_sync_data_with_registrar(auto_commit=False)

            # d1 should be synced with the mocked response.
            self.assertEqual(d1.expiration_date, fields.Date.to_date('2021-08-03'))
            self.assertEqual(d1.delete_date, fields.Date.to_date('2021-10-22'))
            self.assertEqual(d1.email_confirmed, True)
            self.assertEqual(d1.synced, True)

            # d2 is already synced but the email address is not confirmed, so it should be synchronized.
            self.assertEqual(d2.email_confirmed, True)

    def test_renew_date_regression(self):
        """
        Bug description:
        - 1. domain was just renewed
        - 2. expiration date is extended
        - 3. data synchronization starts (depends on cron time)
        - 4. stale data is fetched from the API (expiration date is not yet updated by Gandi)
        - 5. expiration date is set back to the initial value
        - 6. new renew quotation is created the next day

        This test ensures the synchronization only ever updates to later dates.
        """
        db = self._make_one_app_free()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2024-01-01'),  # 2023-01-01 + 1 year
            'delete_date': fields.Date.to_date('2024-03-21'),  # 2024-01-01 + 80 days
            'user_id': self.owner.id,
            'database_id': db.id,
        })

        def _stale_data(*args, **kwargs):
            return {
                'dates': {
                    'hold_begins_at': '2023-01-01T00:00:00Z',
                    'pending_delete_ends_at': '2023-03-22T00:00:00Z',  # 2023-01-01 + 80 days
                },
                'contacts': {
                    'owner': {
                        'reachability': 'done',
                    },
                },
            }

        with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.info', new=_stale_data, create=True):
            self.env['openerp.enterprise.domain']._cron_sync_data_with_registrar(auto_commit=False)
            # The expiration/delete dates should be untouched.
            self.assertEqual(domain.expiration_date, fields.Date.to_date('2024-01-01'))
            self.assertEqual(domain.delete_date, fields.Date.to_date('2024-03-21'))
            # Because the dates don't match, retry later.
            self.assertEqual(domain.synced, False)

        def _up_to_date_data(*args, **kwargs):
            return {
                'dates': {
                    'hold_begins_at': '2024-01-01T00:00:00Z',
                    'pending_delete_ends_at': '2024-03-21T00:00:00Z',
                },
                'contacts': {
                    'owner': {
                        'reachability': 'done',
                    },
                },
            }

        with patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.info', new=_up_to_date_data, create=True):
            self.env['openerp.enterprise.domain']._cron_sync_data_with_registrar(auto_commit=False)
            # The expiration/delete dates should be updated.
            self.assertEqual(domain.expiration_date, fields.Date.to_date('2024-01-01'))
            self.assertEqual(domain.delete_date, fields.Date.to_date('2024-03-21'))  # 2024-01-01 + 80 days
            # Because the dates match, it's fully synced.
            self.assertEqual(domain.synced, True)

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.tag_domain')
    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._route_hostname_to_db')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.put_record')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.create_web_redirection')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.delete_records')
    def test_set_default_dns_config(self, delete_mock, redirection_mock, put_mock, meta_mock, tag_mock):
        delete_mock.return_value = True
        redirection_mock.return_value = {'message': 'OK'}
        put_mock.return_value = ''  # HTTP 201
        meta_mock.return_value = ''
        tag_mock.return_value = ''  # HTTP 201

        db = self._make_one_app_free()
        fqdn = f'{uuid.uuid4()}.com'
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': fqdn,
            'expiration_date': fields.Date.to_date('2021-08-02'),
            'delete_date': fields.Date.to_date('2021-10-20'),
            'user_id': self.owner.id,
            'database_id': db.id,
        })

        self.env['openerp.enterprise.domain']._cron_set_default_dns_config(auto_commit=False)

        self.assertEqual(domain.state, 'integrated')
        delete_mock.assert_called_once_with(fqdn)
        redirection_mock.assert_called_once_with(fqdn, fqdn, f'https://www.{fqdn}', 'http301')
        put_mock.assert_called_once_with(fqdn, 'CNAME', 'www', [f'{db.db_name}.{SAAS_BASE_URL}.'], ttl=None)
        meta_mock.assert_called_once_with(f'www.{fqdn}')
        tag_mock.assert_called_once_with(fqdn, ['saas'])

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.tag_domain')
    @patch('odoo.addons.openerp_enterprise.models.domain.OpenerpEnterpriseDomain._route_hostname_to_db')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.put_record')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.create_web_redirection')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.delete_records')
    def test_set_default_dns_config_failed(self, delete_mock, redirection_mock, put_mock, meta_mock, tag_mock):
        delete_mock.side_effect = Exception('test')
        redirection_mock.return_value = {'message': 'OK'}
        put_mock.return_value = ''  # HTTP 201
        meta_mock.return_value = ''
        tag_mock.return_value = ''  # HTTP 201

        db = self._make_one_app_free()
        fqdn = f'{uuid.uuid4()}.com'
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': fqdn,
            'expiration_date': fields.Date.to_date('2021-08-02'),
            'delete_date': fields.Date.to_date('2021-10-20'),
            'user_id': self.owner.id,
            'database_id': db.id,
        })

        with self.assertLogs('odoo.addons.openerp_enterprise.models.domain', level='ERROR') as cm:
            self.env['openerp.enterprise.domain']._cron_set_default_dns_config(auto_commit=False)
        self.assertIn("failed to delete dns records: domain=" + fqdn, cm.output[0])

        self.assertEqual(domain.state, 'error')

    def test_gc_domains(self):
        db1 = self._make_one_app_free()
        d1 = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.subtract(fields.Date.today(), days=81),
            'delete_date': fields.Date.subtract(fields.Date.today(), days=1),
            'user_id': self.owner.id,
            'database_id': db1.id,
        })

        db2 = self._make_one_app_free()
        d2 = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.subtract(fields.Date.today(), days=81),
            'delete_date': fields.Date.subtract(fields.Date.today(), days=1),
            'user_id': self.owner.id,
            'database_id': db2.id,
            'synced': True,
        })

        self.env['openerp.enterprise.domain']._gc_domains()

        self.assertTrue(d1.active, 'Non-synced record should not be archived.')
        self.assertFalse(d2.active, 'Record should be archived.')

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.resend_reachability_email')
    @freeze_time('2023-01-01')
    def test_resend_confirmation_email_reachability_pending(self, resend_email_mock):
        fqdn = f"{uuid.uuid4()}.com"
        db = self._make_one_app_free()
        self.env['openerp.enterprise.domain'].create({
            'fqdn': fqdn,
            'expiration_date': fields.Date.to_date('2024-01-01'),
            'delete_date': fields.Date.to_date('2024-03-01'),
            'user_id': self.owner.id,
            'database_id': db.id,
            'create_date': fields.Date.today(),
            'synced': True,
        })

        with freeze_time('2023-01-03'):
            with self.subTest('2 days later: email should not be sent'):
                self.env['openerp.enterprise.domain']._cron_resend_confirmation_email()
                resend_email_mock.assert_not_called()
                resend_email_mock.reset_mock()

        with freeze_time('2023-01-06'):
            with self.subTest('5 days later: email should be sent'):
                self.env['openerp.enterprise.domain']._cron_resend_confirmation_email()
                resend_email_mock.assert_called_once_with(fqdn)
                resend_email_mock.reset_mock()

        with freeze_time('2023-01-11'):
            with self.subTest('10 days later: email should be sent'):
                self.env['openerp.enterprise.domain']._cron_resend_confirmation_email()
                resend_email_mock.assert_called_once_with(fqdn)
                resend_email_mock.reset_mock()

        with freeze_time('2023-01-15'):
            with self.subTest('14 days later: email should be sent'):
                self.env['openerp.enterprise.domain']._cron_resend_confirmation_email()
                resend_email_mock.assert_called_once_with(fqdn)
                resend_email_mock.reset_mock()

    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.resend_reachability_email')
    @freeze_time('2023-01-01')
    def test_resend_confirmation_email_already_confirmed(self, resend_email_mock):
        fqdn = f"{uuid.uuid4()}.com"
        db = self._make_one_app_free()
        self.env['openerp.enterprise.domain'].create({
            'fqdn': fqdn,
            'expiration_date': fields.Date.to_date('2021-01-01'),
            'delete_date': fields.Date.to_date('2021-03-01'),
            'user_id': self.owner.id,
            'database_id': db.id,
            'create_date': '2020-01-01',
            'synced': True,
            'email_confirmed': True,
        })

        with freeze_time('2023-01-06'):
            with self.subTest('5 days later: email should not be sent because it is already confirmed'):
                self.env['openerp.enterprise.domain']._cron_resend_confirmation_email()
                resend_email_mock.assert_not_called()

    @freeze_time('2024-01-01')
    @patch('odoo.addons.openerp_enterprise.models.gandi_api.GandiAPI.status')
    def test_send_renewal_reminder(self, status_mock):
        # Setup: create a domain with a renewal quotation.
        status_mock.return_value = {
            'products': [{'prices': [{'price_before_taxes': 10.0}]}],
        }
        self.owner.country_id = self.env.ref('base.be')
        self.env['product.pricelist'].create({
            'name': 'EUROPE EUR',
            'country_group_ids': [fields.Command.link(self.env.ref('base.europe').id)],
            'sequence': 1,
            'currency_id': self.env.ref('base.EUR').id,
        })
        self._subscribe()
        domain = self.env['openerp.enterprise.domain'].create({
            'fqdn': f'{uuid.uuid4()}.com',
            'expiration_date': fields.Date.to_date('2024-01-30'),
            'delete_date': fields.Date.to_date('2024-02-01'),
            'user_id': self.owner.id,
            'database_id': self.db.id,
            'synced': True,
            'email_confirmed': True,
        })
        with self.mock_mail_gateway():
            self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)
            self.assertTrue(domain.renew_order_id, 'sale order not created')
            self.assertEqual(domain.to_renew, False)

        # Move the date forward and verify the reminders are sent.
        for days in [7, 14, 30]:
            date = fields.Date.subtract(domain.expiration_date, days=days)
            with (
                self.subTest(f'reminder D-{days}'),
                freeze_time(date.strftime('%Y-%m-%d')),
                self.mock_mail_gateway(),
            ):
                self.env['openerp.enterprise.domain']._cron_renew_domains(auto_commit=False)
                self.assertMailMailWRecord(
                    domain.renew_order_id,
                    [domain.user_id.partner_id],
                    'sent',
                    content='Renew now',
                    email_values={
                        'email_from': '"OdooBot" <odoobot@example.com>',
                        'subject': f'{domain.fqdn} expires in {days} days',
                    },
                )

@tagged('-standard', 'oe_gandi')
class TestGandiIntegration(TestDNSSetup):
    # These tests make real call (in dry-run mode) to Gandi's API and verify requests are well formatted on Odoo's side.
    # They are not run by default because they are slow and because Gandi is expected to behave as documented anyway.
    # To run these tests the 'oe_gandi' tag will have to be selected explicitely.

    def _register_multi(self, tld, **kwargs):
        for contact_type in [ContactType.company, ContactType.person]:
            with self.subTest(f'{tld}_{contact_type.name}'):
                if contact_type != ContactType.person:
                    kwargs['orgname'] = 'example'

                self._register_dry_run(tld, contact_type, **kwargs)

    def _register_dry_run(self, tld, contact_type, **kwargs):
        assert TEST, 'aborting...'

        domain = uuid.uuid4()
        fqdn = f'{domain}{tld}'

        kwargs.update({
            'type': contact_type.value,
            'given': 'foo',
            'family': 'bar',
            'email': 'foo.bar@example.com',
            'phone': kwargs.get('phone', '0032490000000'),
            'country': kwargs.get('country', 'BE'),
            'streetaddr': 'addr',
            'city': 'city',
            'zip': kwargs.get('zip', '1000'),
        })
        if contact_type != ContactType.person:
            kwargs['orgname'] = 'example'

        response = GANDI.register_domain(fqdn, kwargs, dry_run=True)

        self.assertEqual(response['status'], 'success')

    def test_register(self):
        """Try to register a random domain name under every allowed TLDs.
        Since some TLDs require extra identification data, this test helps detecting missing information.
        Should be run before introducing a new TLD.
        """

        # Special cases have their own tests below, exclude them.
        tlds = [t for t in TLDS if t not in ['.ca', '.es', '.fi', '.gr', '.it', '.se', '.uk', '.co.uk', '.us']]

        for tld in tlds:
            self._register_multi(tld)

    def test_register_ca_cctld(self):
        with self.subTest('Canadian individual'):
            extra_params = {
                'ca_legaltype': 'CCT',
                'country': 'CA',
                'state': 'CA-QC',
                'zip': 'G1L 3N4',
                'phone': '1–604–111–1111',
            }
            self._register_dry_run('.ca', ContactType.person, **extra_params)

        with self.subTest('Canadian company'):
            extra_params = {
                'ca_legaltype': 'CCO',
                'ca_business_entity_type': 'commercial company',
                'country': 'CA',
                'state': 'CA-QC',
                'zip': 'G1L 3N4',
                'phone': '1–604–111–1111',
            }
            self._register_dry_run('.ca', ContactType.company, **extra_params)

    def test_register_es_cctld(self):
        with self.subTest('Spanish individual'):
            extra_params = {
                'es_identification': 'VAT_NUMBER',
                'es_legalform': '1',
                'es_tipo_identification': '0',
                'country': 'ES',
                'state': 'ES-M',
                'zip': '28001',
            }
            self._register_dry_run('.es', ContactType.person, **extra_params)

        with self.subTest('Spanish company'):
            extra_params = {
                'es_identification': 'VAT_NUMBER',
                'es_legalform': '612',
                'es_tipo_identification': '1',
            }
            self._register_dry_run('.es', ContactType.company, **extra_params)

    def test_register_fi_cctld(self):
        with self.subTest('Finn individual'):
            # Are you a Finnish person/organization?
            # If you are a Finnish individual, indicate your Finnish identity number (11 characters).
            extra_params = {
                'fi_isfinish': True,
                'fi_ident_number': '11111111111',
            }
            self._register_dry_run('.fi', ContactType.person, **extra_params)

        with self.subTest('Non Finn individual'):
            # The field birth_date is required.
            extra_params = {
                'fi_isfinish': False,
                'fi_birth_date': '1900-01-01',
            }
            self._register_dry_run('.fi', ContactType.person, **extra_params)

        with self.subTest('Finn business'):
            extra_params = {
                'fi_isfinish': True,
                'fi_business_number': '11111111',
                'fi_organization_type': '1',
            }
            self._register_dry_run('.fi', ContactType.company, **extra_params)

        with self.subTest('Non Finn business'):
            extra_params = {
                'fi_isfinish': False,
                'fi_business_number': '11111111',
                'fi_organization_type': '1',
            }
            self._register_dry_run('.fi', ContactType.company, **extra_params)

    def test_register_gr_cctld(self):
        # .gr requires at least a 2-year registration.
        # {
        #     "object": "OBJECT_DURATION",
        #     "cause": "CAUSE_VALUE",
        #     "code": 400,
        #     "message": "Minimum create duration for example.gr domain is 2 year(s)"
        # }
        with self.subTest('Greek individual'):
            extra_params = {}
            self._register_dry_run('.gr', ContactType.person, **extra_params)

    def test_register_it_cctld(self):
        with self.subTest('Italian individual'):
            # Please be sure you have entered your fiscal code, VAT number or national ID number.
            extra_params = {
                'it_pin': 'ID_NUMBER',
                'country': 'IT',
                'state': 'IT-VE',
                'zip': '30026',
            }
            self._register_dry_run('.it', ContactType.person, **extra_params)

        with self.subTest('Italian company'):
            # The field vat_number is required.
            # Please select your entity type.
            extra_params = {
                'it_registrant_entity_type': '2',
                'it_vat_number': 'VAT_NUMBER',
            }
            self._register_dry_run('.it', ContactType.company, **extra_params)

    def test_register_se_cctld(self):
        with self.subTest('Swedish individual'):
            # The owner of a .SE domain must provide an ID number.\n<br />\n<br />\nIf you are registering it as an individual:\n<ul>\n<li>if you are a Swedish resident, you must provide your <i>personnummer</i>, or if you are registering for a legal entity, its <i>\u201corganisationsnummer</i> in the format XXXXXX-XXXX</li>,\n<li>otherwise, you must provide a personal ID number (drivers' license or passport number, etc.) preceded with the initials of your country of residence in brackets. for example: [GB] )</li>\n</ul>"
            extra_params = {
                'se_ident_number': 'ID_NUMBER',
            }
            self._register_dry_run('.se', ContactType.person, **extra_params)

        with self.subTest('Swedish company'):
            # The field vat_number is required.
            extra_params = {
                'se_ident_number': 'ID_NUMBER',
                'se_registrant_vatid': 'VAT_NUMBER',
            }
            self._register_dry_run('.se', ContactType.company, **extra_params)

    def test_register_uk_cctld(self):
        with self.subTest('UK individual'):
            # Please specify a document type.
            extra_params = {
                'uk_contact_type': 'IND',
            }
            self._register_dry_run('.uk', ContactType.person, **extra_params)

        with self.subTest('UK company'):
            extra_params = {
                'uk_contact_type': 'LTD',
                'uk_co_no': 'COMPANY_NUMBER',
            }
            self._register_dry_run('.uk', ContactType.company, **extra_params)

    def test_register_us_cctld(self):
        with self.subTest('US individual'):
            extra_params = {
                'us_app_purpose': 'P3',
                'us_nexus_category': 'C12',
            }
            self._register_dry_run('.us', ContactType.person, **extra_params)

        with self.subTest('US company'):
            extra_params = {
                'us_app_purpose': 'P5',
                'us_nexus_category': 'C32',
            }
            self._register_dry_run('.us', ContactType.company, **extra_params)

    def test_register_with_bo_address(self):
        # There is no postal code system in Bolivia.
        # Simple regression test for Gandi ticket #15476581.
        # The test should simply pass. It used to throw the following error:
        # {
        #     "object": "OBJECT_CONTACT",
        #     "cause": "CAUSE_BADPARAMETER",
        #     "code": 400,
        #     "message": "Cant associate the required contact with those domains: zip: 'xxx' is not valid for BO"
        # }
        extra_params = {'country': 'BO'}
        self._register_dry_run('.com', ContactType.person, **extra_params)
